// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2010 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Net;

using Smdn.Net.Pop3.Client.Session;
using Smdn.Net.Pop3.Client.Transaction;
using Smdn.Net.Pop3.Protocol.Client;
using Smdn.Net.MessageAccessProtocols;

namespace Smdn.Net.Pop3.Client {
  public class PopClient : IDisposable {
    public PopClientProfile Profile {
      get { return profile; }
    }

    public int Timeout {
      get { return (session == null) ? profile.Timeout : session.TransactionTimeout; }
      set
      {
        profile.Timeout = value;

        if (session != null)
          session.TransactionTimeout = value;
      }
    }

    public int SendTimeout {
      get { return (session == null) ? profile.SendTimeout : session.SendTimeout; }
      set
      {
        profile.SendTimeout = value;

        if (session != null)
          session.SendTimeout = value;
      }
    }

    public int ReceiveTimeout {
      get { return (session == null) ? profile.ReceiveTimeout : session.ReceiveTimeout; }
      set
      {
        profile.ReceiveTimeout = value;

        if (session != null)
          session.ReceiveTimeout = value;
      }
    }

    public bool IsConnected {
      get { return (session != null && session.State != PopSessionState.NotConnected); }
    }

    public bool DeleteAfterRetrieve {
      get; set;
    }

    internal PopSession Session {
      get { CheckConnected(); return session; }
    }

    public PopCapabilityList ServerCapabilities {
      get { return Session.ServerCapabilities; }
    }

    /*
     * construction
     */
    public PopClient()
    {
      profile = new PopClientProfile();
    }

    public PopClient(Uri authority)
    {
      profile = new PopClientProfile(authority);
    }

    public PopClient(string host, string userName)
      : this()
    {
      profile.Host = host;
      profile.UserName = userName;
    }

    public PopClient(string host, int port, string userName)
      : this()
    {
      profile.Host = host;
      profile.Port = port;
      profile.UserName = userName;
    }

    public PopClient(string host, int port, string userName, string authType)
      : this()
    {
      profile.Host = host;
      profile.Port = port;
      profile.UserName = userName;
    }

    public void Connect(string password)
    {
      Connect(password, null);
    }

    public void Connect(string password, UpgradeConnectionStreamCallback createSslStreamCallback)
    {
      Connect(new NetworkCredential(profile.UserName, password), createSslStreamCallback);
    }

    public void Connect(ICredentialsByHost credentials)
    {
      Connect(credentials, null);
    }

    public void Connect(ICredentialsByHost credentials, UpgradeConnectionStreamCallback createSslStreamCallback)
    {
      try {
        profile.SetCredentials(credentials);

        ThrowIfError(PopSessionCreator.CreateSession(profile,
                                                     createSslStreamCallback ?? PopConnection.CreateSslStream,
                                                     out session));

        Refresh();
      }
      finally {
        profile.SetCredentials(null);
      }
    }

    void IDisposable.Dispose()
    {
      Disconnect(false);
    }

    public void Disconnect()
    {
      Disconnect(false);
    }

    public void Logout()
    {
      Disconnect(true);
    }

    public void Disconnect(bool logout)
    {
      if (session != null)
        session.Disconnect(logout);

      session = null;
    }

    private PopSession session;
    private PopClientProfile profile;

    /*
     * operations
     */
    public long MessageCount {
      get { return dropListing.MessageCount; }
    }

    public long TotalSize {
      get { return dropListing.SizeInOctets; }
    }

    private PopDropListing dropListing;
    private Dictionary<long, PopMessageInfo> messages = new Dictionary<long, PopMessageInfo>();

    public void Refresh()
    {
      ThrowIfError(Session.Stat(out dropListing));
    }

    public IEnumerable<PopMessageInfo> GetMessages()
    {
      PopScanListing[] scanLists;

      ThrowIfError(Session.List(out scanLists));

      foreach (var scanList in scanLists) {
        yield return ToMessage(scanList);
      }
    }

    public PopMessageInfo GetMessage(long messageNumber)
    {
      PopScanListing scanList;

      ThrowIfError(Session.List(messageNumber, out scanList));

      return ToMessage(scanList);
    }

    private PopMessageInfo ToMessage(PopScanListing scanList)
    {
      var message = new PopMessageInfo(this, scanList);

      messages[message.MessageNumber] = message;

      return message;
    }

    public void CancelDelete()
    {
      ThrowIfError(Session.Rset());

      lock (((System.Collections.ICollection)messages).SyncRoot) {
        foreach (var message in messages.Values) {
          message.Reset();
        }
      }
    }

    public void NoOp()
    {
      ThrowIfError(Session.NoOp());
    }

    internal static void ThrowIfError(PopTransactionResult result)
    {
      if (result.Failed)
        throw new PopException(result.Description);
    }

    /*
     * methods for checking and validating internal status
     */
    private void CheckConnected()
    {
      if (!IsConnected)
        throw new InvalidOperationException("not connected or disconnected");
    }
  }
}
